// ******************************************************************
//
// File Name   : PB_io.c
// Author      : D. Festing
// Created     : 21/7/2010
// Revised     : 
// Version     : 1.0
// Target MCU  : AVR ATmega32
// Editor Tabs	: 3
//
// Description : This file contains all initialisation
//               for the AVR IO ports and peripherals 
//
// Notes       : "Lazy refresh" of IO ports
//               void RED_LED_ON(void)
//               {
//                  RED_LED_DDR  |= (1<<RED_LED_DATA_BIT);
//                  RED_LED_PORT |= (1<<RED_LED_DATA_BIT);
//               }
//
//               check for correct transitions
//
// ******************************************************************


#include "project.h" // has to be the first header file

#include "usart.h"
#include "Spi.h"
#include "adc.h"


// ******************************************************************
//
// Function Name:     initialise_hardware
// Description:       initialises the hardware to the state required 
//                    for this project 
// Preconditions:     None
// Input Parameters:  None
// Return Parameter:  None
// Effects:           
//
// Notes:             This function should be done with interrupts
//                    disabled.
//
// ******************************************************************
void initialise_hardware(void)
{
// On all ports without pull-up or pull-down resistors, either set the port to
// input pull-up enabled or output low, don't leave them floating!
// even though the JEDMicro dev board has pull-ups, enabled the port pull-ups as well.

// Port initialisation
// Inputs - Tri-state DDxn = 0 and PORTxn = 0
// Inputs - Pull-up   DDxn = 0 and PORTxn = 1
// Outputs - DDxn = 1 and PORTxn = 0

// PortA, PA0, PA1, PA2, PA2 and PA4 ADC, all others input pull-up enabled
   DDRA = 0x00;  // 0000 0000
   PORTA = 0xE0; // 1110 0000
   
// PORTB, PB0 (DAC select) and PB3 output low, all others input pull-up enabled
// N.B. you should always ensure that PB4 is not floating (_SS)
//      probably a good idea to make sure PB6 is pull-up enabled
// PB2 is connected to output of RTC chip
// PB4, PB5 and PB7 get configured as outputs later on (SS, MOSI and SCK) 
   DDRB = 0x09;  // 0000 1001
	PORTB = 0xF6; // 1111 0110
   
// PortC, PC0 and PC1 input (I2C with hardwired pullups),
// all anothers output low which is inverted by the FETs
   DDRC = 0xFC;  // 1111 1100
   PORTC = 0x00; // 0000 0000

// PortD, PD0 and PD1 RS232 (reset values), PD2, PD3 and PD6
// input pull up enabled and PD4, PD5 and PD7 output low   
   DDRD = 0xB0;  // 1011 0000
   PORTD = 0x4C; // 0100 1100

// UART initialisation
// Communication Parameters: 8 Data, 1 Stop, No Parity
// UART Transmitter: On
// UART Baud rate: 9600

   #define USART_BAUDRATE 9600
   #define BAUD_PRESCALE (((F_CPU / (USART_BAUDRATE * 16UL))) - 1)

   UBRRL = BAUD_PRESCALE; // Load lower 8-bits of the baud rate value into the low byte of the UBRR register
   UBRRH = (BAUD_PRESCALE >> 8); // Load upper 8-bits of the baud rate value into the high byte of the UBRR
                                  // register
                                   
   UCSRA = (1<<TXC); // need to clear this flag or you get a USART_TXC_vect when
                      // global interrupts are enabled (in osSTART).  This is a result
                      // of using the bootloader (MegaLoad) that came with the
                      // ATmega32 JEDmicro dev board.  The original MegaLoad also has
                      // this problem (curtvm).
   UCSRB |= ((1<<RXCIE) | (1<<RXEN) | (1<<TXCIE) | (1<<TXEN)); 
//   UCSRC |= (1<<URSEL) | (1<<UCSZ0) | (1<<UCSZ1); // Use 8-bit character sizes - URSEL
                                                    // bit set to select the UCRSC register 
// set by default to 8N1
   
   usartInit(); // software initialisation of USART parameters

   adc_init();
   SpiInit();
}
/* ---------- end of hardware_initialise() ---------- */

// ******************************************************************
//
// Function Names:    see below
// Description:       initialises the output ports to the state required
//                    every time the function is called, ie "lazy refresh" 
// Preconditions:     None
// Input Parameters:  None
// Return Parameter:  None
// Effects:           None
//
// Notes:             
//
// ******************************************************************

void RED_LED_ON(void)
{
   RED_LED_DDR  |= (1<<RED_LED_DATA_BIT);
   RED_LED_PORT |= (1<<RED_LED_DATA_BIT);
}

void RED_LED_OFF(void)
{
   RED_LED_DDR  |= (1<<RED_LED_DATA_BIT);
   RED_LED_PORT &= ~(1<<RED_LED_DATA_BIT);  
}

void RED_LED_TOGGLE(void)
{
   RED_LED_DDR  |= (1<<RED_LED_DATA_BIT);
   RED_LED_PORT ^= (1<<RED_LED_DATA_BIT);
}

void ERROR_FLASH(uint8_t error_code)
{
   int16_t i;

   
   RED_LED_DDR  |= (1<<RED_LED_DATA_BIT);
   
   FOREVER
   {
      RED_LED_PORT ^= (1<<RED_LED_DATA_BIT);
//      _delay_ms(50); // max 70ms
      for(i = 0; i < 30000; i++)
         {;}

      while(!(UCSRA & (1 << UDRE)));
      UDR = error_code;   
   }   
}
