// ******************************************************************
//
// File Name   : process_serial_comms.c
// Author      : Robert Glassey
// Created     : 27/3/2006
// Revised     : 28/6/2008, Dave Festing to NEW DATA FORMAT
//               and ported to the Atmel AVR ATmega series
// Version     : 1.0
// Target MCU  : Atmel AVR series
// Editor Tabs	: 3
//
// Description : This file contains serial send and receive functions
//
// Notes: 
//   - presents command_length, command_buffer and sci_cmd_send to external
//     functions
//
//   - when a valid command is received and command buffer is free
//     (command_length = 0), the new command is copied into the command buffer
//     and command_length becomes the length of the new command.
//
//   - command_length must be cleared by the external functions to free the
//     command buffer for new commands. This should be done as soon as the
//     command string has been processed and before any other functions are
//     executed to avoid buffer overruns should a new command become available.
//
//   - to transmit a command string - call
//     sci_cmd_send(char cmd, char s[], char len_s);
//     passing in the command, byte string and length of string
//
//   - cannot test for UART hardware errors here. This must be done before reading UDR
//     therefore this has to happen in uart.c
//
// ******************************************************************


#include "project.h"     // has to be the first header file

#include "process_serial_comms.h"
#include "usart.h"


static volatile uint8_t command_length = 0; // >0 when command available, cleared when
                                            // processed. The flag that passes control
                                            // from the rx_func() back to main()
static volatile uint8_t command_buffer[COMMS_PROTOCOL_BUFFER_LENGTH] = {0};
                     // holds a copy of rx_buffer[]
enum{NO_ERRORS = 0, DATA_ERROR = 1, PARITY_ERROR = 2, DATA_OVERRUN = 3, FRAME_ERROR = 4,
     COMMAND_BUFFER_NOT_READY = 5, CRC_ERROR = 6, BUFFER_OVERFLOW_ERROR = 7} rx_error;


void process_RX_data(uint8_t received_byte)
{
   static uint8_t rx_buffer_index = 0;
   static uint8_t rx_crc = 0;
   static uint8_t rx_buffer[COMMS_PROTOCOL_BUFFER_LENGTH] = {'\0'};
// incoming command string before CRC checking
// these variables static as information needs to be retained between function calls 
   int i  = 0;


   if(received_byte & 0x80) // 1XXX XXXX = command
   {
      rx_buffer_index = 0;
      rx_buffer[rx_buffer_index++] = (uint8_t)(received_byte & 0x7F);
      rx_crc = addcrc(received_byte, 0);
      rx_error = NO_ERRORS;
   }
   else if(received_byte & 0x40) // 01XX XXXX = CRC and end of string
   {
      if(((rx_crc & 0x3F) == (received_byte & 0x3F)) && (rx_error == NO_ERRORS))

   // CRC's match and no errors, copy rx_buffer[] into command_buffer[]
      {
         if(command_length == 0) // check command buffer ready
         {
            for(i = 0; i < rx_buffer_index; i++)
            {
               command_buffer[i] = rx_buffer[i];
            }

            command_length = rx_buffer_index;
         }
         else
         {
            rx_error = COMMAND_BUFFER_NOT_READY;
         }    
      }
      else
      {
         rx_error = CRC_ERROR;
      }
   }
   else if(received_byte & 0x20) // you definitely got data
   {
         
      rx_crc = addcrc(received_byte, rx_crc); // add byte to crc 
      
      if(received_byte & 0x10)
      {
         rx_buffer[rx_buffer_index++] |= (uint8_t)(received_byte << 4);
      // 0011 HHHH, HIGH NIBBLE, increment index, OR into rx_buffer[] 
      }
      else
      {
         rx_buffer[rx_buffer_index] = (uint8_t)(received_byte & 0x0F);
      // 0010 LLLL, LOW NIBBLE
      }
   
      if(rx_buffer_index >= COMMS_PROTOCOL_BUFFER_LENGTH)
      {
         rx_error = BUFFER_OVERFLOW_ERROR;
      }
   }
   else
   {
      rx_error = DATA_ERROR;
   }
   
   if((rx_error == CRC_ERROR) || (rx_error == COMMAND_BUFFER_NOT_READY) || (rx_error == DATA_ERROR) || (rx_error == BUFFER_OVERFLOW_ERROR)) 
   {
      ; // put error flash into here
   }
   else
      {;}
}
/* ---------- end of ISR(USART_RXC_vect) ---------- */


uint8_t addcrc(uint8_t added_data, uint8_t crc)
{
   uint8_t i = 0;


   for(i = 0; i < 8; i++)
   {
      crc = (uint8_t)((crc << 1) | (((crc >> 7) ^ (crc >> 1) ^ crc ^ added_data) & 1));
      added_data >>= 1;
   }

   return crc;
}
/* ---------- end of addcrc() ---------- */


void sci_cmd_send(uint8_t command, const uint8_t data[], uint8_t number_of_data_bytes)
{
   uint8_t raw_data = 0;
   uint8_t data_to_send = 0;
   uint8_t crc = 0;
   int8_t i = 0;

   
   data_to_send = (uint8_t)(command | 0x80);            // COMMAND = 1XXX XXX

   usartSendByte(data_to_send);
   
   crc = addcrc(data_to_send, crc);

   for(i = 0; i < number_of_data_bytes; i++)
   {
      raw_data = data[i];
      data_to_send = (uint8_t)(0x20|(raw_data & 0x0F));  // 0010 LLLL LOW NIBBLE

      usartSendByte(data_to_send);
      
      crc = addcrc(data_to_send, crc); 
      raw_data >>= 4;
      data_to_send = (uint8_t)(0x30|(raw_data & 0x0F));  // 0011 HHHH HIGH NIBBLE

      usartSendByte(data_to_send);
      
      crc = addcrc(data_to_send, crc);
   }

   data_to_send = (uint8_t)(0x40 | (crc & 0x3F)); // CRC = 01XX XXXX

   usartSendByte(data_to_send);
}
/* ---------- end of sci_cmd_send() ---------- */


uint8_t get_command_length(void)
{
   return command_length;
}   
/* ---------- end of get_command_length() ---------- */


void reset_command_length(void)
{
   command_length = 0;
}   
/* ---------- end of reset_command_length() ---------- */


uint8_t get_command_buffer(uint8_t index)
{
   uint8_t command_buffer_content = 0;
   
   
   command_buffer_content = command_buffer[index];
   
   return command_buffer_content; // returns a copy of command_buffer_content
}   
/* ---------- end of get_command_buffer() ---------- */
